import numpy as np
from .constraints import MODE_NUMBER, MODE_ALPHANUMERIC, MODE_BYTE, MODE_KANJI, MODE_ECI, MODE_MIXED, ERROR_CORRECTION_LEVEL_L, ERROR_CORRECTION_LEVEL_M, ERROR_CORRECTION_LEVEL_Q, ERROR_CORRECTION_LEVEL_H, EC_CODEWORDS, FONT_SIZE_LARGE_LARGE, FONT_SIZE_LARGE, FONT_SIZE_SMALL, FONT_SIZE_MEDIUM, POSITION_BOTTOM_LEFT, POSITION_BOTTOM_RIGHT, POSITION_TOP_LEFT, POSITION_TOP_RIGHT, POSITION_MIDDLE
from .errorCorrection import ErrorCorrection
from .dataConverter import DataConverter

# TODO: cache all the module_sequence for each version
class GenerateQR:
    def generate(self, data, version = None, error_correction = ERROR_CORRECTION_LEVEL_L):
        encoded_data, version = self.get_encoded_data(data, version, error_correction)
        codewords = self.get_codewords(encoded_data, version, error_correction)
        module_sequence = self.get_module_sequence(codewords, version, error_correction)
        optimanl_mask = self.get_optimal_mask(codewords ,module_sequence, version, error_correction)
        return optimanl_mask, version

    def print_qr(self, matrix):
        # char_on = '█'  # Or you can use '1'
        char_on = '#'
        char_off = ' ' # Or you can use '0'
        for row in matrix:
            row_str = ''.join(char_on if cell == 1 else char_off for cell in row)
            print(row_str)
        print()

    def get_encoded_data(self, data, version, error_correction):
        dataConverter = DataConverter(version, error_correction)
        encoded_data = dataConverter.encode(data)
        return encoded_data, dataConverter.version

    def get_codewords(self, encoded_data, version, error_correction):
        table = EC_CODEWORDS[version][error_correction] 
        blocks = []
        ec_blocks = []
        codewords = []

        for i in range(table[2] + table[4]):
            if i < table[2]:
                block_size = table[3]
            else:
                block_size = table[5]
            blocks.append(encoded_data[:block_size])
            encoded_data = encoded_data[block_size:]
            ec_blocks.append(ErrorCorrection().getEDC(blocks[i], block_size+table[1]))

        for c in range(table[3]):
            for j in range(table[2] + table[4]):
                codewords.append(blocks[j][c])
        for j in range(table[2], table[2] + table[4]):
            codewords.append(blocks[j][table[5] - 1])
        for c in range(table[1]):
            for j in range(table[2]+table[4]):
                codewords.append(ec_blocks[j][c])
        
        return codewords

    def get_module_sequence(self, codewords, version, error_correction):
        size = 17 + 4 * version
        matrix = np.zeros((size, size), dtype=int)

        self._fill_area(matrix, 0, 0, 9, 9)  # Top-left finder pattern
        self._fill_area(matrix, 0, size - 8, 8, 9)  # Top-right finder pattern
        self._fill_area(matrix, size - 8, 0, 9, 8)  # Bottom-left finder pattern

        # Alignment patterns
        alignment_tracks = GenerateQR.get_alignment_coordinates(version)
        last_track = len(alignment_tracks) - 1 
        for row_index, row in enumerate(alignment_tracks):
            for column_index, column in enumerate(alignment_tracks):
                if (row_index == 0 and (column_index == 0 or column_index == last_track)) or \
                    (column_index == 0 and row_index == last_track):
                    continue
                self._fill_area(matrix, row - 2, column - 2, 5, 5)
        
        # Timing patterns
        self._fill_area(matrix, 6, 9, version * 4, 1)  # Horizontal timing pattern
        self._fill_area(matrix, 9, 6, 1, version * 4)  # Vertical timing pattern

        # Dark module
        matrix[size - 8, 8] = 1

        # Version information
        if version >= 7:
            self._fill_area(matrix, 0,size - 11, 3, 6)
            self._fill_area(matrix, size - 11, 0, 6, 3)

        row_step = -1
        row = size - 1
        column = size - 1
        sequence = []
        index = 0

        # elaborete the insertion sequence of the modules
        while column >= 0:
            if matrix[row, column] == 0:
                sequence.append((row, column))
            
            if index % 2 == 1:
                row += row_step
                if row == -1 or row == size:
                    row_step = -row_step
                    row += row_step
                    column -= 2 if column == 7 else 1
                else:
                    column += 1
            else:
                column -= 1
            
            index += 1
        return sequence

    @staticmethod
    def get_alignment_coordinates(version):
        if version == 1:
            return []
        intervals = version // 7 + 1
        distance = 4 * version + 4
        step = -(-distance // (intervals * 2)) * 2  
        return [6] + [distance + 6 - (intervals - 1 - index) * step for index in range(intervals)]

    def _fill_area(self, matrix ,row, column, width, height, fill=1):
        fill_row = [fill] * width
        matrix[row:row + height, column:column + width] = fill_row

    def get_optimal_mask(self, codewords, module_sequence, version, error_correction):
        best_matrix = None
        best_score = float('inf')
        for index in range(8):
            matrix = self.get_mask_matrix(index, codewords, module_sequence, version, error_correction) 
            penalty_score = self.get_penalty_score(matrix) 
            if penalty_score < best_score:
                best_score = penalty_score
                best_matrix = matrix
        return best_matrix

    def get_mask_matrix(self, mask_index, codewords, module_sequence, version, error_correction):
        MASK_FNS = [
            lambda row, column: ((row + column) & 1) == 0,
            lambda row, column: (row & 1) == 0,
            lambda row, column: column % 3 == 0,
            lambda row, column: (row + column) % 3 == 0,
            lambda row, column: (((row >> 1) + column // 3) & 1) == 0,
            lambda row, column: ((row * column) & 1) + ((row * column) % 3) == 0,
            lambda row, column: ((((row * column) & 1) + ((row * column) % 3)) & 1) == 0,
            lambda row, column: ((((row + column) & 1) + ((row * column) % 3)) & 1) == 0,
        ]
        size = version * 4 + 17
        # apply codewords and mask
        mask_fn = MASK_FNS[mask_index]
        matrix = np.zeros((size, size), dtype=int)
        for index, (row, column) in enumerate(module_sequence):
            if index >> 3 >= len(codewords):
                break
            # Each codeword contains 8 modules, so shifting the index to the right by 3 gives the codeword's index
            codeword = codewords[index >> 3]
            bit_shift = 7 - (index & 7)
            module_bit = (codeword >> bit_shift) & 1
            # Apply the mask function to the module bit
            matrix[row][column] = module_bit ^ int(mask_fn(row, column))

        # calculate the format information
        FORMAT_DIVISOR = np.array([1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1], dtype=int)
        FORMAT_MASK = np.array([1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0], dtype=int)
        format_poly = np.zeros(15, dtype=int)
        if error_correction == ERROR_CORRECTION_LEVEL_L:
            error_index = 1
        elif error_correction == ERROR_CORRECTION_LEVEL_M:
            error_index = 0
        elif error_correction == ERROR_CORRECTION_LEVEL_Q:
            error_index = 3
        elif error_correction == ERROR_CORRECTION_LEVEL_H:
            error_index = 2
        format_poly[0] = error_index >> 1
        format_poly[1] = error_index & 1
        format_poly[2] = mask_index >> 2
        format_poly[3] = (mask_index >> 1) & 1
        format_poly[4] = mask_index & 1
        rest = ErrorCorrection().poly_rest(format_poly, FORMAT_DIVISOR)
        format_poly[5:] = rest
        masked_format_poly = np.bitwise_xor(format_poly, FORMAT_MASK)
        format_modules = masked_format_poly.tolist()

        # place format information
        # Top-left
        matrix[8, :6] = format_modules[:6]
        matrix[8, 7:9] = format_modules[6:8]
        matrix[8, -8:] = format_modules[7:]
        matrix[7, 8] = format_modules[8]
        # Bottom-left
        for index in range(7):
            matrix[size - index - 1, 8] = format_modules[index]
        # Top-right
        for index in range(6):
            matrix[5 - index, 8] = format_modules[9 + index]
        
        # place fixed patterns
        # Finder patterns
        for row, col in [(0, 0), (size - 7, 0), (0, size - 7)]:
            self._fill_area(matrix, row, col, 7, 7)
            self._fill_area(matrix, row + 1, col + 1, 5, 5, fill=0)
            self._fill_area(matrix, row + 2, col + 2, 3, 3)
        
        # Separators
        self._fill_area(matrix, 7, 0, 8, 1, fill=0)
        self._fill_area(matrix, 0, 7, 1, 7, fill=0)
        self._fill_area(matrix, size - 8, 0, 8, 1, fill=0)
        self._fill_area(matrix, 0, size - 8, 1, 7, fill=0)
        self._fill_area(matrix, 7, size - 8, 8, 1, fill=0)
        self._fill_area(matrix, size - 7, 7, 1, 7, fill=0)

        # Alignment pattern
        alignment_tracks = GenerateQR.get_alignment_coordinates(version)
        last_track = len(alignment_tracks) - 1

        for row_index, row in enumerate(alignment_tracks):
            for column_index, column in enumerate(alignment_tracks):
                # Skipping the alignment near the finder patterns
                if (row_index == 0 and (column_index == 0 or column_index == last_track)) or \
                    (column_index == 0 and row_index == last_track):
                    continue
                
                # Fill a 5x5 area with `1`
                self._fill_area(matrix, row - 2, column - 2, 5, 5)
                # Clear a 3x3 area with `0`
                self._fill_area(matrix, row - 1, column - 1, 3, 3, 0)
                # Set the specific cell to `1`
                matrix[row][column] = 1

        # Timing patterns
        for pos in range(8, size - 8, 2):
            matrix[6, pos] = 1
            matrix[6, pos + 1] = 0
            matrix[pos, 6] = 1
            matrix[pos + 1, 6] = 0
        matrix[6, size - 7] = 1
        matrix[size - 7, 6] = 1

        # Dark module
        matrix[size - 8, 8] = 1
        # version information
        if version < 7:
            return matrix
        version_info = self.get_version_information(version)
        for index, bit in enumerate(version_info):
            row = index // 3
            col = index % 3
            # Update matrix at specific positions
            matrix[5 - row][size - 9 - col] = bit
            matrix[size - 11 + col][row] = bit
        return matrix

    def get_version_information(self, version):
        VERSION_DIVISOR = [1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1]
        version_bin_str = format(version, '06b') + '000000000000'
        poly = [int(b) for b in version_bin_str]
        # Perform polynomial division
        poly_rest_result = ErrorCorrection().poly_rest(poly, VERSION_DIVISOR)
        return poly[:6] + poly_rest_result

    def get_penalty_score(self, matrix):
        RULE_3_PATTERN = np.array([1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0], dtype=np.uint8)
        RULE_3_REVERSED_PATTERN = RULE_3_PATTERN[::-1]
        total_penalty = 0
        
        # Rule 1
        row_penalty = sum(self.get_line_penalty(row) for row in matrix)
        total_penalty += row_penalty
        
        column_penalty = sum(self.get_line_penalty(matrix[:, column_index]) for column_index in range(matrix.shape[1]))
        total_penalty += column_penalty
        
        # Rule 2
        blocks = 0
        size = matrix.shape[0]
        for row in range(size - 1):
            for column in range(size - 1):
                module = matrix[row, column]
                if (matrix[row, column + 1] == module and
                    matrix[row + 1, column] == module and
                    matrix[row + 1, column + 1] == module):
                    blocks += 1
        total_penalty += blocks * 3
        
        # Rule 3
        patterns = 0
        for index in range(size):
            row = matrix[index]
            for column_index in range(size - 11):
                if any(np.array_equal(pattern, row[column_index:column_index + 11]) for pattern in [RULE_3_PATTERN, RULE_3_REVERSED_PATTERN]):
                    patterns += 1
                    
            for row_index in range(size - 11):
                if any(np.array_equal(pattern, matrix[row_index:row_index + 11, index]) for pattern in [RULE_3_PATTERN, RULE_3_REVERSED_PATTERN]):
                    patterns += 1
        total_penalty += patterns * 40
        
        # Rule 4
        total_modules = size * size
        dark_modules = np.sum(matrix)
        percentage = dark_modules * 100 / total_modules
        mix_penalty = abs(int(percentage / 5 - 10)) * 10
        
        return total_penalty + mix_penalty

    def get_line_penalty(self, line):
        count = 0
        counting = None
        penalty = 0
        for cell in line:
            if cell != counting:
                counting = cell
                count = 1
            else:
                count += 1
                if count == 5:
                    penalty += 3
                elif count > 5:
                    penalty += 1
        return penalty


class ErrorCorrection:
    def __init__(self, ):
        self.prim = 2
        self.exp = 8
        self.irreducible_poly=285
        self.size = self.prim ** self.exp
        self.LOG = bytearray(self.size)
        self.EXP = bytearray(self.size)
        self._generate_tables()
    
    def _generate_tables(self):
        value = 1
        for i in range(self.size - 1):  # Iterate over the range of the field size
            self.EXP[i] = value
            self.LOG[value] = i

            # Multiply `value` by 2 in GF(2^exp) field
            value <<= 1
            # Reduce `value` modulo irreducible polynomial if necessary
            if value >= self.size:
                value ^= self.irreducible_poly

        # Special case for `value = 0`
        self.LOG[0] = 0
        self.EXP[self.size - 1] = 1  # EXP[255] = 1 for GF(2^8)

    # in GF(2^n) the addition/sottraction is the XOR operation -> a = -a
    def add(self, a, b):
        if a > self.size or b > self.size:
            raise ValueError("The value is out of the field")
        return a ^ b
    
    def mul(self,a,b):
        if a == 0 or b == 0:
            return 0
        return self.EXP[(self.LOG[a] + self.LOG[b]) % 255]
    
    def div(self,a,b):
        if b == 0:
            raise ZeroDivisionError("Division by zero")
        if a == 0:
            return 0
        return self.EXP[(self.LOG[a] + (255 - self.LOG[b])) % 255]
    
    def poly_mul(self, poly1, poly2):
        len1 = len(poly1)
        len2 = len(poly2)
        coeffs = [0] * (len1 + len2 - 1)
        
        for i in range(len1):
            for j in range(len2):
                coeffs[i + j] ^= self.mul(poly1[i], poly2[j])

        return coeffs
    
    def poly_rest(self, dividend, divisor):
        rest = list(dividend)
        while len(rest) >= len(divisor):
            if rest[0] != 0:
                factor = self.div(rest[0], divisor[0])
                subtr = self.poly_mul([factor] + [0] * (len(rest) - len(divisor)), divisor)
                rest = [rest[i] ^ subtr[i] for i in range(len(rest))]
            rest.pop(0)
        return rest
    
    def get_generator_poly(self, degree):
        last_poly = [1]

        for _ in range(degree):
            last_poly = self.poly_mul(last_poly, [1, self.EXP[_]])

        return last_poly
    
    def getEDC(self, data, codewords):
        degree = codewords - len(data)
        messagePoly = bytearray(codewords)
        messagePoly[:len(data)] = data
        generatorPoly = self.get_generator_poly(degree)
        remainder = self.poly_rest(messagePoly, generatorPoly)
        return list(remainder)



from .constraints import MODE_NUMBER, MODE_ALPHANUMERIC, MODE_BYTE, MODE_KANJI, EC_CODEWORDS, ERROR_CORRECTION_LEVEL_M, ERROR_CORRECTION_LEVEL_L, ERROR_CORRECTION_LEVEL_Q, ERROR_CORRECTION_LEVEL_H
from urllib.parse import urlparse
import math

class DataConverter:        
    def __init__(self, version = None, error_correction = ERROR_CORRECTION_LEVEL_L) -> None:
        self._buffer = ""
        self.version = version
        self.error_correction = error_correction
        self.VALUE_GEN_MAP = {
            MODE_NUMBER: self._encode_numeric,
            MODE_ALPHANUMERIC: self._encode_alphanumeric,
            MODE_BYTE: self._encode_byte,
            MODE_KANJI: self._encode_kanji
        }

    def encode(self, data: str):
        # detect mode 
        data, encoding_mode = self._detect_mode(data)

        # encode data
        data_generator = self.VALUE_GEN_MAP[encoding_mode]
        data_generator(data)

        # determine version if not provided and raise error if data is too long
        if self.version is None:
            count_codewords = math.ceil(len(self._buffer) / 8)
            for version in range(1, 41):
                if EC_CODEWORDS[version][self.error_correction][0] >= count_codewords:
                    self.version = version
                    break
            if self.version is None:
                raise ValueError("Data is too long")
        elif self.version > 40 or self.version < 1:
            raise ValueError("Invalid version number")
        elif math.ceil(len(self._buffer) / 8) > EC_CODEWORDS[self.version][self.error_correction][0]:
            raise ValueError("Data is too long")
               
        # add mode indicator and character count indicator
        CHAR_COUNT_INDICATOR = {
            MODE_NUMBER: [10, 12, 14],
            MODE_ALPHANUMERIC: [9, 11, 13],
            MODE_BYTE: [8, 16, 16],
            MODE_KANJI: [8, 10, 12]
        }
        if self.version < 10:
            length_bits =  CHAR_COUNT_INDICATOR[encoding_mode][0]
        elif self.version < 27:
            length_bits = CHAR_COUNT_INDICATOR[encoding_mode][1]
        else:
            length_bits = CHAR_COUNT_INDICATOR[encoding_mode][2]
        self._add_onhead_buffer(len(data), length_bits)
        self._add_onhead_buffer(encoding_mode, 4)

        # add terminator
        self._add_buffer(0, 4)

        # add filler
        byte = int(len(self._buffer) / 8)
        for i in range(EC_CODEWORDS[self.version][self.error_correction][0] - byte):
            self._add_buffer(0b11101100 if i % 2 == 0 else 0b00010001, 8)

        return [int(self._buffer[i:i+8], 2) for i in range(0, len(self._buffer), 8)]
    
    def get_total_data_codewords(self):
        return EC_CODEWORDS[self.version][self.error_correction][0]
    
    def get_total_codewords(self):
        return EC_CODEWORDS[self.version][self.error_correction][0] + EC_CODEWORDS[self.version][self.error_correction][1]*(EC_CODEWORDS[self.version][self.error_correction][2]+EC_CODEWORDS[self.version][self.error_correction][4])
    
    def get_ec_codewords_per_block(self):
        return EC_CODEWORDS[self.version][self.error_correction][1]

    def get_version(self):
        return self.version

    def _detect_mode(self, data):
        if (urlparse(data).scheme in ('http', 'https') and urlparse(data).netloc != ''):
            data = data.upper()

        is_numeric = all(char.isdigit() for char in data)
        is_alphanumeric = all(char in "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:" for char in data)
        if is_numeric:
            return data, MODE_NUMBER
        elif is_alphanumeric:
            return data, MODE_ALPHANUMERIC
        elif self._is_kanji_mode(data):
            return data, MODE_KANJI
        else:
            return data, MODE_BYTE

    def _is_kanji_mode(self, data):
        try:
            sjis_encoded = data.encode('shift-jis')
        except UnicodeEncodeError:
            return False
        
        i = 0
        while i < len(sjis_encoded):
            if sjis_encoded[i:i+1] == b' ':     
                i += 1
                continue
            if i + 1 >= len(sjis_encoded):
                return False
            code = (sjis_encoded[i] << 8) + sjis_encoded[i + 1]
            if (0x8140 <= code <= 0x9FFC) or (0xE040 <= code <= 0xEBBF):
                i += 2
            else:
                return False
        return True
    
    def _add_buffer(self, value:int, bit_length:int):
        self._buffer += f"{value:0{bit_length}b}"

    def _add_onhead_buffer(self, value:int, bit_length:int):
        self._buffer = f"{value:0{bit_length}b}" + self._buffer

    def _encode_numeric(self, data):
        for i in range(0, len(data), 3):
            chunk = data[i:i+3]
            if len(chunk) == 3:
                self._add_buffer(int(chunk), 10)
            elif len(chunk) == 2:
                self._add_buffer(int(chunk), 7)
            else:
                self._add_buffer(int(chunk), 4)

    def _encode_alphanumeric(self, data):
        ALPHANUMERIC_TABLE = {
            '0': 0, '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7, '8': 8, '9': 9,
            'A': 10, 'B': 11, 'C': 12, 'D': 13, 'E': 14, 'F': 15, 'G': 16, 'H': 17, 'I': 18, 'J': 19,
            'K': 20, 'L': 21, 'M': 22, 'N': 23, 'O': 24, 'P': 25, 'Q': 26, 'R': 27, 'S': 28, 'T': 29,
            'U': 30, 'V': 31, 'W': 32, 'X': 33, 'Y': 34, 'Z': 35, ' ': 36, '$': 37, '%': 38, '*': 39,
            '+': 40, '-': 41, '.': 42, '/': 43, ':': 44
        }
        for i in range(0, len(data), 2):
            chunk = data[i:i + 2]
            if len(chunk) == 2:
                self._add_buffer(ALPHANUMERIC_TABLE[chunk[0]]*45 + ALPHANUMERIC_TABLE[chunk[1]], 11)
            elif len(chunk) == 1:
                self._add_buffer(ALPHANUMERIC_TABLE[chunk], 6)

    def _encode_byte(self, data: str):
        if self._is_latin1(data):
            for char in data:
                self._add_buffer(ord(char), 8)
        else:
            for byte in data.encode('utf-8'):
                self._add_buffer(byte, 8)

    def _is_latin1(self, char : str) -> bool:
        try:
            char.encode('latin-1')
        except UnicodeEncodeError:
            return False
        return True 
    
    def _encode_kanji(self, data: str):
        for char in data:
            sjis_encoded = char.encode('shift-jis')

            siji_value = (sjis_encoded[0] << 8) + sjis_encoded[1]
            if 0x8140 <= siji_value <= 0x9FFC:
                msb = sjis_encoded[0] - 0x81
                lsb = sjis_encoded[1] - 0x40
            elif 0xE040 <= siji_value <= 0xEBBF:
                msb = sjis_encoded[0] - 0xC1
                lsb = sjis_encoded[1] - 0x40
            else:
                raise ValueError(f"Invalid Shift-JIS value: {siji_value}")

            self._add_buffer((msb * 0xC0) + lsb, 13)



from .custom_qr import CustomQR
from .generate_qr import GenerateQR
from .constraints import ERROR_CORRECTION_LEVEL_H, ERROR_CORRECTION_LEVEL_Q, ERROR_CORRECTION_LEVEL_M, ERROR_CORRECTION_LEVEL_L
import cv2

class QrCode:
    """
    QrCode class to generate and print QR code
    """	
    def __init__(self):
        self.qr = GenerateQR()
        self.custom = CustomQR()
    
    def generate(self, data, version=None, error_correction=ERROR_CORRECTION_LEVEL_L):
        return self.qr.generate(data, version, error_correction)
    
    def create_qr_image(self, matrix, 
                        background=(255,255,255), 
                        block_style={"size": 10, "type" : 0, "color":[(0,0,0)]}, 
                        finder_style={"color":(0,0,0)},
                        alignment_style = None):
        return self.custom.draw_qr(matrix, background, block_style, finder_style, alignment_style)
    
    def print_qr_console(self, matrix):
        self.qr.print_qr(matrix)

    def display_qr(self, img):
        cv2.imshow('image', img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    def create_image_file(self, img, filename="qr.png"):
        cv2.imwrite(filename, img)
    
    def write_text(self, img, text, 
                    text_style={"color":(0,0,0), "size": "small", "bot": 0, "left": 0, "orientation": 0}, 
                    background=(255,255,255), block_size=10):
        return self.custom.write_text(img, text_style, background, block_size, text)


# from .qr_code import QrCode
from .qrCode import QrCode
from .constraints import (
    MODE_NUMBER, MODE_ALPHANUMERIC, MODE_BYTE, MODE_KANJI, MODE_ECI,
    MODE_MIXED, ERROR_CORRECTION_LEVEL_L, ERROR_CORRECTION_LEVEL_M,
    ERROR_CORRECTION_LEVEL_Q, ERROR_CORRECTION_LEVEL_H, EC_CODEWORDS
)



import cv2
import numpy as np
from .generate_qr import GenerateQR
from .charset import small

class CustomQR:
    """
    Draw a QR code with custom blocks, custom finder, custom  and text

    draw_qr(matrix, background, block_style, finder_style, alignment_style): 
        Draw a QR code with custom blocks, custom finder, custom alignment and custom text
    
    write_text(img, text_style, background, block_size, text):
        Write a text on the QR code

    struct block_style:
        size: int
        type: int       # 0: square, 1: circle
        color: list of tuples
    
    struct finder_style:
        color: tuple
    
    struct alignment_style:
        color: tuple
    
    struct text_style:
        color: tuple
        size: str    # "small"
        bot: int
        left: int
        orientation: int
    """	
    def draw_qr(self, matrix, 
                background = (255,255,255), 
                block_style = {"size": 10, "type" : 0, "color":[(0,0,0)]}, 
                finder_style = {"color":(0,0,0)}, 
                alignment_style = None):
        # if alignment_style is not defined, use the finder style
        if alignment_style is None:
            alignment_style = finder_style

        block_size = block_style["size"]
        real_size = len(matrix) * block_size
        matrix_size = len(matrix)

        img = np.full((real_size, real_size, 3), background, dtype=np.uint8) # Create the image 
        
        img = self._draw_blocks(img, matrix, block_style, background) # Draw the modules/blocks
        self._draw_finder(img, matrix_size, finder_style, block_size, background) # Draw the finder patterns
        self._draw_alignment(img, matrix_size, alignment_style, block_size, background) # Draw the alignment patterns

        return img
    
    def write_text(self, img, text_style, background, block_size, text):
        if not self.check_space(img, text, block_size, text_style):
            return None # not enough space
        
        color = self._rgb_to_bgr(text_style["color"])
        size = text_style["size"]
        bot = text_style["bot"]
        left = text_style["left"]
        orientation = text_style["orientation"]

        if size == "small":
            charset = small
        if orientation == 1:
            text = text[::-1]
        for i, character in enumerate(text):
            if character in charset:
                char = charset[character]
                box_width = self.write_char(img, char, bot, left, block_size, color, background, orientation)
                if orientation == 0:
                    left += box_width
                elif orientation == 1:
                    left -= box_width
        return img
    
    def write_char(self, img, char, bot, left, block_size, color, background, orientation):
        box_height, box_width = char.shape

        box = np.full((box_height * block_size, box_width * block_size, 3), background, dtype=np.uint8)
        color_block = np.full((block_size, block_size, 3), color, dtype=np.uint8)
        for i in range(char.shape[0]):
            for j in range(char.shape[1]):
                if char[i][j] == 1:
                    box[i*block_size:(i+1)*block_size, j*block_size:(j+1)*block_size] = color_block


        w_margin = np.full((block_size, box_width * block_size, 3), background, dtype=np.uint8)
        if bot > 0:
            # add the margin to the bottom of the box
            box = np.concatenate((box, w_margin), axis=0)
            box_height += 1 # update the height of the box with the margin
            bot -= 1
        if (bot + box_height) * block_size < img.shape[0]:
            # add the margin to the top of the box
            box = np.concatenate((w_margin, box), axis=0)
            box_height += 1
        

        h_margin = np.full((box_height * block_size, block_size, 3), background, dtype=np.uint8)
        
        if orientation == 0:
            if left > 0:
                box = np.concatenate((h_margin, box), axis=1)
                box_width += 1
                left -= 1 
            if (left + box_width) * block_size < img.shape[1]:
                box = np.concatenate((box, h_margin), axis=1)
                box_width += 1
            img_left_start = left * block_size
            img_left_end = img_left_start + box.shape[1]
        elif orientation == 1:
            right = left - box_width 
            if right > 0:
                box = np.concatenate((h_margin,box), axis=1)
                box_width += 1
                right -= 1
            if left > 0:
                box = np.concatenate((box, h_margin), axis=1)
                box_width += 1
                left += 1
            img_left_end = left * block_size
            img_left_start = right * block_size

        img_bot_start = img.shape[0] - (bot * block_size) - box.shape[0]
        img_bot_end = img.shape[0] - (bot * block_size)
        img[img_bot_start:img_bot_end, img_left_start:img_left_end] = box

        return char.shape[1] + 1 # add 1 to consider the space between characters   

    def check_space(self, img, text, box_size, text_style):
        size = text_style["size"]
        bot = text_style["bot"]
        left = text_style["left"]
        orientation = text_style["orientation"]

        matrix_len = int(img.shape[0] // box_size)

        if orientation == 0:
            margin = left
        elif orientation == 1:
            margin = matrix_len - left

        total_width = 0
        total_height = 0
        if size == "small":
            for i, c in enumerate(text):
                if c in small:
                    char_height, char_width = small[c].shape
                    # if i == 0 and left != 0:
                    #     total_width += left
                    if i == 0 and margin != 0:
                        total_width += margin
                    if i != 0 : # sum the space between characters
                        total_width += 1
                    total_width += char_width
                    
                    total_height = max (total_height, char_height)

        if orientation == 0:
            return matrix_len - (total_height + bot) >= 0 and matrix_len - (total_width + left) >= 0
        elif orientation == 1:
            return matrix_len - (total_height + bot) >= 0 and left - total_width >= 0

    def _rgb_to_bgr(self, color):
        return (color[2], color[1], color[0])

    def _draw_blocks(self, img, matrix, block_style, background):
        if block_style["type"] == 0: # square
            self._draw_blocks_square(img, matrix, block_style)
        elif block_style["type"] == 1: # circle
            self._draw_blocks_circle(img, matrix, block_style, background)
        return img

    def _draw_blocks_square(self, img, matrix, block_style):
        block_size = block_style["size"]
        colors = block_style["color"]
        color_len = len(colors)

        colored_block = np.array(self._rgb_to_bgr(colors[0]), dtype=np.uint8) if color_len == 1 else None

        for i in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    if matrix[i][j] == 1:
                        if color_len > 1:
                            color = self._rgb_to_bgr(colors[(i * matrix.shape[1] + j) % color_len])
                            colored_block = np.array(color, dtype=np.uint8)
                        img[i*block_size:(i+1)*block_size, j*block_size:(j+1)*block_size] = colored_block
    
    def _draw_blocks_circle(self, img, matrix, block_style, background):
        block_size = block_style["size"]
        colors = block_style["color"]
        color_len = len(colors)

        radius = int(block_size // 2.1)

        # Pre-generate the colored circles
        colored_circles = []
        for color in colors:
            color = self._rgb_to_bgr(color)
            colored_circle = np.full((block_size, block_size, 3), background, dtype=np.uint8)
            colored_circle = cv2.circle(colored_circle, (block_size//2, block_size //2), radius, color, -1, cv2.LINE_AA)
            colored_circles.append(colored_circle)

        for i in range(matrix.shape[0]):
            for j in range(matrix.shape[1]):
                if matrix[i][j] == 1:
                    color_index = (i * matrix.shape[1] + j) % color_len
                    top_left_x = j * block_size
                    top_left_y = i * block_size
                    img[top_left_y:top_left_y + block_size, top_left_x:top_left_x + block_size] = colored_circles[color_index]
    
    def _draw_finder(self, img, length, finder_style, block_size, background):
        color = self._rgb_to_bgr(finder_style['color'])
        finder_positions = [(0,0), (0, length-7), (length-7, 0)]

        finder = np.full((7*block_size, 7*block_size, 3), color, dtype=np.uint8)
        finder[block_size: 6*block_size, block_size: 6* block_size] = background
        finder[2*block_size: 5*block_size, 2*block_size: 5*block_size] = color

        for row, col in finder_positions:
            top_left_x = row * block_size
            top_left_y = col * block_size
            bottom_right_x = (row + 7) * block_size
            bottom_right_y = (col + 7) * block_size

            img[top_left_y:bottom_right_y, top_left_x:bottom_right_x] = finder
    
    def _draw_alignment(self, img, length, alignment_style, block_size, background):
        color = self._rgb_to_bgr(alignment_style['color'])

        alignment_tracks = GenerateQR.get_alignment_coordinates(int((length-21)/4)+1)
        last_track = len(alignment_tracks) - 1

        alignment = np.full ((5*block_size, 5*block_size, 3), color, dtype=np.uint8)
        alignment[block_size:4*block_size, block_size:4*block_size] = background
        alignment[2*block_size:3*block_size, 2*block_size:3*block_size] = color

        for row_index, row in enumerate(alignment_tracks):
            for column_index, column in enumerate(alignment_tracks):
                # Skipping the alignment near the finder patterns
                if (row_index == 0 and (column_index == 0 or column_index == last_track)) or \
                   (column_index == 0 and row_index == last_track):
                    continue
                
                top_left_x = (column - 2)* block_size
                top_left_y = (row - 2) * block_size
                bottom_right_x = (column + 5 - 2) * block_size
                bottom_right_y = (row + 5 - 2) * block_size

                img[top_left_y:bottom_right_y, top_left_x:bottom_right_x] = alignment
                




from setuptools import setup, find_packages

setup(
    name='custom_qr',
    version='0.9.4',
    author='Daniele Ye',
    author_email='daniele.ye03@gmail.com',
    description='Generate and customize QR codes',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/DaYe03/custom_qr.git',
    packages=find_packages(),
    install_requires=['numpy', 'opencv-python'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)
